//#include <stdio.h>
//#include <stdint.h>
#include <vmelib.h>
#include <CAENVMEtypes.h>
#include <CAENVMElib.h>
//#include <numpy/arrayobject.h>



using namespace std;


CAENVMEDevice::CAENVMEDevice( unsigned char bd_type,
                              unsigned char link,
                              unsigned char bd_num,
                              unsigned char addr_mode,
                              unsigned char data_width,
                              unsigned char verbosity,
                              unsigned int base_addr )
{
   this->bd_type = bd_type;
   this->link = link;
   this->bd_num = bd_num;
   this->addr_mode = addr_mode;
   this->data_width = data_width;
   this->verbosity = verbosity;
   // verbosity = 0: no messages
   // verbosity = 1: only error messages
   // verbosity = 2: error and diagnostic messages
   if (this->verbosity==2)
      printf("\tInitializing device...\n");
   this->open_board((CVBoardTypes) bd_type, link, bd_num);
   this->open_device(base_addr, addr_mode, data_width);
}

///////////////////////////////
CAENVMEDevice::~CAENVMEDevice()
{
   int err_code;
   err_code = CAENVME_End(this->handle);

   if (this->verbosity==2 || err_code)
      printf("\tClose communication board: error %s\n", this->decode_error(err_code));

   if (this->verbosity==2)
      printf("\tDevice deleted.\n");
}

/////////////////////////////////////////////////////
int CAENVMEDevice::open_board( unsigned char bd_type,
                               unsigned char link,
                               unsigned char bd_num )
{
   this->bd_type = bd_type;
   this->link = link;
   this->bd_num = bd_num;
   int err_code;

   err_code = CAENVME_End(this->handle);  // try to close

  if (this->verbosity==2)
      printf("\tTry to close communication board: %s\n", this->decode_error( err_code));

   err_code = CAENVME_Init((CVBoardTypes) bd_type, this->bd_num, this->link, &(this->handle));

   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\tOpening communication board: %s\n", this->decode_error( err_code));
   if (this->verbosity==2)
      printf("\t\tbd_type: %x,\t\tlink: %x,\t\tbd_num: %x\n", bd_type, link, bd_num);
   return err_code;
}


//////////////////////////////////////////////////////////
int CAENVMEDevice::open_device( unsigned int base_addr,
                                unsigned char addr_mode,
                                unsigned char data_width )
{
   this->base_addr = base_addr;
   if (addr_mode!=INVALID_MODE)    this->addr_mode  = addr_mode;
   if (data_width!=INVALID_MODE)   this->data_width = data_width;
   if (this->verbosity==2) {
      printf("\tSetting device parameters:\n");
      print_params(this->addr_mode, this->data_width, this->base_addr);
    }
   return cvSuccess;
}


// read/write functions

//////////////////////////////////////////////////////
int CAENVMEDevice::read_reg( unsigned int reg_addr,
                             unsigned int base_addr,
                             unsigned char addr_mode,
                             unsigned char data_width)
{
   int err_code;
   unsigned int addr = base_addr!=INVALID_BA ? reg_addr + base_addr : reg_addr + this->base_addr;
   if (this->verbosity==2) {
      printf("\tReading register 0x%x:\n", addr);
      print_params(addr_mode!=INVALID_MODE ? addr_mode : this->addr_mode,
                   data_width!=INVALID_MODE ? data_width : this->data_width,
                   base_addr);
   }
   if((data_width==cvD64 || (!data_width && this->data_width==cvD64)) && this->verbosity){
      printf("\tCan't execute a D64 Read Cycle\n");
      return cvInvalidParam;
   }

   this->single_buffer = 0;
   err_code = CAENVME_ReadCycle( this->handle, addr, &(this->single_buffer),
                                 (CVAddressModifier) (addr_mode!=INVALID_MODE ? addr_mode : this->addr_mode),
                                 (CVDataWidth)      (data_width!=INVALID_MODE ? data_width : this->data_width));

   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));
   return err_code;
}


////////////////////////////////////////////////////////
int CAENVMEDevice::write_reg( unsigned int data,
                              unsigned int reg_addr,
                              unsigned int base_addr,
                              unsigned char addr_mode,
                              unsigned char data_width )
{
   int err_code = 0;
   unsigned int addr;
   
   addr = base_addr!=INVALID_BA ? reg_addr + base_addr : reg_addr + this->base_addr;
   if (this->verbosity==2) {
      printf("\tWriting 0x%x to register 0x%x:\n", data, addr);
      print_params(addr_mode!=INVALID_MODE  ? addr_mode  : this->addr_mode,
                   data_width!=INVALID_MODE ? data_width : this->data_width,
                   base_addr);
   }
   if((data_width==cvD64 || (!data_width && this->data_width==cvD64)) && this->verbosity){
      printf("\tCan't execute a D64 Write Cycle\n");
      return cvInvalidParam;
   }

   err_code = CAENVME_WriteCycle( this->handle, addr, &data,
                                  (CVAddressModifier) (addr_mode!=INVALID_MODE ? addr_mode : this->addr_mode),
                                  (CVDataWidth) (data_width!=INVALID_MODE ? data_width : this->data_width) );

   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));

   return err_code;
}


/////////////////////////////////////////////////////////
int CAENVMEDevice::read_write( unsigned int data,
                               unsigned int reg_addr,
                               unsigned int base_addr,
                               unsigned char addr_mode,
                               unsigned char data_width )
{
   int err_code = 0;
   unsigned int addr;
   
   addr = base_addr!=INVALID_BA ? reg_addr + base_addr : reg_addr + this->base_addr;
   if (this->verbosity==2) {
      printf("\tReading and subsequent writing 0x%x to register 0x%x:\n", data, addr);
      print_params(addr_mode!=INVALID_MODE  ? addr_mode:  this->addr_mode,
                   data_width!=INVALID_MODE ? data_width: this->data_width,
                   base_addr);
   }
   if((data_width==cvD64 || (!data_width && this->data_width==cvD64)) && this->verbosity){
      printf("\tCan't execute a D64 Write Cycle\n");
      return cvInvalidParam;
   }

   err_code = CAENVME_RMWCycle( this->handle, addr, &data,
                                  (CVAddressModifier) (addr_mode!=INVALID_MODE ? addr_mode : this->addr_mode),
                                  (CVDataWidth) (data_width!=INVALID_MODE ? data_width : this->data_width) );
   this->single_buffer = data;
   
   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));

   return err_code;
}


///////////////////////////////////////////////////////
int CAENVMEDevice::read_page( unsigned int size,
                              unsigned int reg_addr,
                              unsigned int base_addr,
                              unsigned char addr_mode,
                              unsigned char data_width,
                              bool  blt)
{
   unsigned char new_addr_mode, new_data_width;
   int err_code;
   unsigned int addr;
   addr = base_addr!=INVALID_BA ? reg_addr + base_addr : reg_addr + this->base_addr;
   new_addr_mode = addr_mode!=INVALID_MODE ? addr_mode : this->addr_mode;
   new_data_width = data_width!=INVALID_MODE ? data_width : this->data_width;

   if (this->verbosity==2)
      printf("\tReading %d registers from address 0x%x:\n", size, addr);

   if(new_addr_mode==cvA16_U) {
      printf(" Can't execute a A16 BLT or Multi Read Cycle\n");
      return cvInvalidParam;
   }
   if(new_addr_mode==cvCR_CSR) {
      printf(" Can't execute a CR/CSR BLT or Multi Read Cycle\n");
      return cvInvalidParam;
   }
   
   /*
   if (new_data_width == cvD64) {
      if (new_addr_mode==cvA24_U_DATA)
         new_addr_mode = cvA24_U_MBLT;
      else
         new_addr_mode = cvA32_U_MBLT;
   }
   else {
      if (new_addr_mode==cvA24_U_DATA)
         new_addr_mode = cvA24_U_BLT;
      else 
         new_addr_mode = cvA32_U_BLT;
   }
//   */
   if (this->verbosity==2)
      print_params(new_addr_mode, new_data_width, base_addr);

   for (unsigned int j=0; j < size; j++){
      this->blt_buffer[j] = 0;
   }
   if (blt) {
      // real BLT cycle
      if (this->verbosity==2)
         printf("Performing a VME block transfer read cycle.\n");
      int n_bytes_read;
      err_code = CAENVME_BLTReadCycle( this->handle, addr, (char *) (this->blt_buffer),
                                       size * (new_data_width & 0x7),
                                       (CVAddressModifier) new_addr_mode,
                                       (CVDataWidth) new_data_width,
                                        &n_bytes_read);
      this->single_buffer = n_bytes_read / (new_data_width & 0x7);
   }
   else {
      // many single vme cycle
      if (this->verbosity==2)
         printf("Performing a block of single VME write cycles.\n");
      unsigned int * addrs = new unsigned int [size];
      CVAddressModifier * addr_modes = new CVAddressModifier [size];
      CVDataWidth * data_widths = new CVDataWidth [size];
      CVErrorCodes * err_codes = new CVErrorCodes [size];
      for (unsigned int i=0; i<size; i++) {
         addrs[i] = addr + i * (new_data_width & 0x7);
         addr_modes[i] = (CVAddressModifier) new_addr_mode;
         data_widths[i] = (CVDataWidth) new_data_width;
         err_codes[i] = (CVErrorCodes) -666;
      }
      err_code = CAENVME_MultiRead( this->handle, addrs, this->blt_buffer, size,
                                    addr_modes, data_widths, err_codes);
      this->single_buffer = 0;
      for (unsigned int i=0; i<size; i++) {
         this->single_buffer += (err_codes[i] == 0) ? 1 : 0;
         err_code |= err_codes[i];
      }
   }
   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));
   return err_code;
}


////////////////////////////////////////////////////////
int CAENVMEDevice::write_page( unsigned int data [],
                               unsigned int size,
                               unsigned int reg_addr,
                               unsigned int base_addr,
                               unsigned char addr_mode,
                               unsigned char data_width,
                               bool  blt)
{
   unsigned char new_addr_mode, new_data_width;
   int err_code;
   unsigned int addr;
   addr = base_addr!=INVALID_BA ? reg_addr + base_addr : reg_addr + this->base_addr;
   new_addr_mode = addr_mode!=INVALID_MODE ? addr_mode : this->addr_mode;
   new_data_width = data_width!=INVALID_MODE ? data_width : this->data_width;

   if (this->verbosity==2)
      printf("\tWriting %d registers from address 0x%x:\n", size, addr);

   if(new_addr_mode==cvA16_U) {
      printf(" Can't execute a A16 BLT or Multi Write cycle\n");
      return cvInvalidParam;
   }
   if(new_addr_mode==cvCR_CSR) {
      printf(" Can't execute a CR/CSR BLT or Multi Write cycle\n");
      return cvInvalidParam;
   }

   if (this->verbosity==2)
      print_params(new_addr_mode, new_data_width, base_addr);

   for (unsigned int j=0; j < size; j++){
      this->blt_buffer[j] = 0;
      printf("%d\t", data[j]);
   }
   printf("\n");
   if (blt) {
      // real BLT cycle
      if (this->verbosity==2)
         printf("Performing a VME block transfer write cycle.\n");
      int n_bytes_written;
      err_code = CAENVME_BLTWriteCycle( this->handle, addr, (void *) data,
                                        size * (new_data_width & 0x7),
                                        (CVAddressModifier) new_addr_mode,
                                        (CVDataWidth) new_data_width,
                                        &n_bytes_written);

      this->single_buffer = n_bytes_written / (new_data_width & 0x7);
   }
   else {
      // many single vme cycle
      if (this->verbosity==2)
         printf("Performing a block of single VME write cycles.\n");
      unsigned int * addrs = new unsigned int [size];
      CVAddressModifier * addr_modes = new CVAddressModifier [size];
      CVDataWidth * data_widths = new CVDataWidth [size];
      CVErrorCodes * err_codes = new CVErrorCodes [size];
      for (unsigned int i=0; i<size; i++) {
         addrs[i] = addr + i * (new_data_width & 0x7);
         addr_modes[i] = (CVAddressModifier) addr_mode;
         data_widths[i] = (CVDataWidth) data_width;
      }
      err_code = CAENVME_MultiWrite( this->handle, addrs, data, size,
                                     addr_modes, data_widths, err_codes);
      printf("err_code = %d\n", err_code);
      this->single_buffer = 0;
      for (unsigned int i=0; i<size; i++) {
         this->single_buffer += (err_codes[i] == 0) ? 1 : 0;
         err_code |= err_codes[i];
      }
   }
   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));
   return err_code;
}


////////////////////////////////////////////////////////////
int CAENVMEDevice::read_multi( unsigned int size,
                               unsigned int regs [],
                               unsigned char addr_modes [],
                               unsigned char data_widths [])
{
   int err_code;


   unsigned int * addrs = new unsigned int [size];
   CVErrorCodes * err_codes = new CVErrorCodes [size];
   CVAddressModifier * ams = new CVAddressModifier [size];
   CVDataWidth * dws = new CVDataWidth [size];

   for (unsigned int j=0; j < size; j++){
      addrs[j] = regs[j]>0xffff ? regs[j] : regs[j] + this->base_addr;
      if(addrs[j]==cvA16_U) {
         printf(" Can't execute a A16 BLT Read Cycle\n");
         return cvInvalidParam;
      }
      ams[j] = (CVAddressModifier) (addr_modes[j]!=INVALID_MODE  ? addr_modes[j]  : this->addr_mode);
      if(ams[j]==cvCR_CSR) {
         printf(" Can't execute a CR/CSR BLT Read Cycle\n");
         return cvInvalidParam;
      }
      dws[j] = (CVDataWidth)       (data_widths[j]!=INVALID_MODE ? data_widths[j] : this->data_width);
   }
   if (this->verbosity==2) {
      printf("\tReading %d registers:\t0x%x\n", size, addrs[0]);
      for (unsigned int i=1; i<size; i++)
         printf(",\t%x", addrs[i]);
      printf("\n");
   }
   // many single vme cycle
   if (this->verbosity==2)
      printf("Performing a block of single VME write cycles.\n");
   err_code = CAENVME_MultiRead( this->handle, addrs, this->blt_buffer, size,
                                 ams, dws, err_codes);
   this->single_buffer = 0;
   for (unsigned int i=0; i<size; i++) {
      this->single_buffer += (err_codes[i] == 0) ? 1 : 0;
      err_code |= err_codes[i];
   }
   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));
   return err_code;
}


////////////////////////////////////////////////////////////
int CAENVMEDevice::write_multi(unsigned int data [],
                               unsigned int size,
                               unsigned int regs [],
                               unsigned char addr_modes [],
                               unsigned char data_widths [])
{
   int err_code;


   
   unsigned int * addrs = new unsigned int [size];
   CVErrorCodes * err_codes = new CVErrorCodes [size];
   CVAddressModifier * ams = new CVAddressModifier [size];
   CVDataWidth * dws = new CVDataWidth [size];

   for (unsigned int j=0; j < size; j++){
      addrs[j] = regs[j]>0xffff ? regs[j] : regs[j] + this->base_addr;
      if(addrs[j]==cvA16_U) {
          printf(" Can't execute a A16 Multi Write\n");
          return cvInvalidParam;
      }
      ams[j] = (CVAddressModifier) (addr_modes[j]!=INVALID_MODE  ? addr_modes[j]  : this->addr_mode);
      if(ams[j]==cvCR_CSR) {
         printf(" Can't execute a CR/CSR Multi Write\n");
         return cvInvalidParam;
      }
      dws[j] = (CVDataWidth)       (data_widths[j]!=INVALID_MODE ? data_widths[j] : this->data_width);
   }
   if (this->verbosity==2) {
      printf("\tWriting %d registers:\t0x%x\n", size, addrs[0]);
      for (unsigned int i=1; i<size; i++)
         printf(", %x", addrs[i]);
      printf("\n");
   }
   // many single vme cycle
   if (this->verbosity==2)
      printf("Performing a block of single VME write cycles.\n");
   err_code = CAENVME_MultiWrite(this->handle, addrs, data, size,
                                 ams, dws, err_codes);
   this->single_buffer = 0;
   for (unsigned int i=0; i<size; i++) {
      this->single_buffer += (err_codes[i] == 0) ? 1 : 0;
      err_code |= err_codes[i];
   }
   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));
   return err_code;
}


//////////////////////////////////////////////////
int CAENVMEDevice::irq_enable(  unsigned int mask)
{
   if (this->verbosity==2)
      printf("\tEnabling IRQ, mask = %s:\n", bin(mask));

   int err_code = CAENVME_IRQEnable( this->handle, mask);

   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));
   return err_code;
}
   

//////////////////////////////////////////////////
int CAENVMEDevice::irq_disable( unsigned int mask)
{
   if (this->verbosity==2)
      printf("\tDisabling IRQ, mask = %s:\n", bin(mask));

   int err_code = CAENVME_IRQDisable( this->handle, mask);

   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));

   return err_code;
}
   

/////////////////////////////////////////////////
int CAENVMEDevice::irq_wait(unsigned int mask,
                            unsigned int timeout)
{
   if (this->verbosity==2)
      printf("\tWaiting IRQ, mask = %s, timeout = %u:\n", bin(mask), timeout);

   int err_code = CAENVME_IRQWait( this->handle, mask, timeout);

   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));

   return err_code;
}

   
//////////////////////////////
int CAENVMEDevice::irq_check()
{
   unsigned char mask;
   if (this->verbosity==2)
      printf("\tChecking IRQ...");

   int err_code = CAENVME_IRQCheck(this->handle, &mask);

   this->single_buffer = mask;

   if (this->verbosity==2)
      printf("\tReturned IRQ mask: %s \n", bin(mask));
 
   if ((this->verbosity>=1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));

   return err_code;
}


/*
///////////////////////////////////////////////////////////////////////
int CAENVMEDevice::reset()
{
   if (this->verbosity==2)
      printf("\tResetting device...\n");

   int err_code = CAENVME_DeviceReset(this->handle);
   printf("SystemReset: %s\n", this->decode_error(err_code));

   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));

   return err_code;
}
*/

///////////////////////////////////////////////////////////////////////
int CAENVMEDevice::get_timeout()
{
   CVVMETimeouts tm;
//   int err_code = -1;
   if (this->verbosity==2)
      printf("\tReading VME bus timeout...\n");

   int err_code = CAENVME_GetTimeout(this->handle, &tm);
   this->single_buffer = tm > 0 ? 400 : 50; // us
   
   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));

   return err_code;
}


///////////////////////////////////////////
int CAENVMEDevice::set_timeout(int timeout)
{
   char tm = timeout < 150 ? 0 : 1;
   if (this->verbosity==2)
      printf("\tWriting VME bus timeout...\n");

   int err_code = CAENVME_SetTimeout(this->handle, (CVVMETimeouts) tm);

   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));

   return err_code;
}


/////////////////////////////////////
int CAENVMEDevice::get_arbiter_type()
{
   CVArbiterTypes arbiter_type;

   if (this->verbosity==2)
      printf("\tReading arbiter type...\n");

   int err_code = CAENVME_GetArbiterType(this->handle, &arbiter_type);
   this->single_buffer = arbiter_type;
   
   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));

   return err_code;
}


//////////////////////////////////////////////////////
int CAENVMEDevice::set_arbiter_type(char arbiter_type)
{
   if (this->verbosity==2)
      printf("\tWriting arbiter type...\n");
   int err_code = CAENVME_SetArbiterType(this->handle, (CVArbiterTypes) arbiter_type);

   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));

   return err_code;
}


//////////////////////////////////
int CAENVMEDevice::get_requester_type()
{
   CVRequesterTypes requester_type;

   if (this->verbosity==2)
      printf("\tReading bus requester type...\n");

   int err_code = CAENVME_GetRequesterType(this->handle, &requester_type);

   this->single_buffer = requester_type;
   
   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));

   return err_code;
}


////////////////////////////////////////////////
int CAENVMEDevice::set_requester_type(char requester_type)
{
   if (this->verbosity==2)
      printf("\tWriting bus requester type...\n");
   
   int err_code = CAENVME_SetRequesterType(this->handle, (CVRequesterTypes) requester_type);
   
   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));

   return err_code;
}


//////////////////////////////////
int CAENVMEDevice::get_requester_level()
{
   CVBusReqLevels requester_level;

   if (this->verbosity==2)
      printf("\tReading bus requester priority level...\n");

   int err_code = CAENVME_GetBusReqLevel(this->handle, &requester_level);
   this->single_buffer = requester_level;
   
   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));

   return err_code;
}


////////////////////////////////////////////////
int CAENVMEDevice::set_requester_level(char requester_level)
{
   if (this->verbosity==2)
      printf("\tWriting bus requester priority level...\n");
   
   int err_code = CAENVME_SetBusReqLevel(this->handle, (CVBusReqLevels) requester_level);
   
   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));

   return err_code;
}


//////////////////////////////////
int CAENVMEDevice::get_release_type()
{
   CVReleaseTypes release_type;

   if (this->verbosity==2)
      printf("\tReading bus release type...\n");

   int err_code = CAENVME_GetReleaseType(this->handle, &release_type);

   this->single_buffer = release_type;
   
   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));

   return err_code;
}


////////////////////////////////////////////////
int CAENVMEDevice::set_release_type(char release_type)
{
   if (this->verbosity==2)
      printf("\tWriting bus release type...\n");
   
   int err_code = CAENVME_SetReleaseType(this->handle, (CVReleaseTypes) release_type);
   
   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));

   return err_code;
}


////////////////////////////////
int CAENVMEDevice::get_sw_info()
{
   if (this->verbosity==2)
      printf("\tRequesting software info...\n");

   char sw_rel[20], bd_rel[20];
   int err_code = CAENVME_SWRelease(sw_rel);
   if (this->verbosity==2)
      printf("\t\tSwRel: %s,\t%s\n", sw_rel, this->decode_error(err_code));

   err_code |= CAENVME_BoardFWRelease(this->handle, bd_rel);
   sprintf(this->string_buffer, "%s | %s", sw_rel, bd_rel);
   if (this->verbosity==2)
      printf("\t\tBdRel: %s,\t%s\n", bd_rel, this->decode_error(err_code));
   
//   err_code = CAENVME_DriverRelease(this->handle, ver);
   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));

   return err_code;
}


/////////////////////////
int CAENVMEDevice::test()
{
   int err_code = -1;
   printf("\tTEST FUNCTION\n");
   /*
   unsigned char page [264];
   for (unsigned int j=0; j < 264; j++)
      page[j] = 17;
   
   FILE * outf = fopen("flash_page.dat", "wb");
   for (int i=0; i<1240; i++) {
       err_code = CAENVME_ReadFlashPage(this->handle, page, i);
       for (unsigned int j=0; j < 264; j++)
          fprintf(outf, "%c", page[j]);
   }
   fclose(outf);
//   printf("ReadFlash: %s\n", this->decode_error(err_code));
//   printf("Buffer:\n");
    */

//    err_code = CAENVME_SetTimeout(int32_t Handle, CVVMETimeouts Value);
    /*              The function sets the specified VME bus timeout on the module. */
   unsigned int addr = 0x32101800;
//   unsigned int data = 0x4880;
   short data = 0xff;
   int count;
//   unsigned int data = 0x0;

//   err_code = CAENVME_GetFIFOMode(this->handle, &data);
   err_code = CAENVME_SetFIFOMode(this->handle, 1);
   printf("fifo: %d (err %d)", data, err_code);

   /*
   CVBusReqLevels requester_level;
   err_code = CAENVME_GetBusReqLevel(this->handle, &requester_level);
   this->single_buffer = requester_level;
   */

   if ((this->verbosity==1 && err_code) || this->verbosity==2)
      printf("\t\t%s\n", this->decode_error(err_code));
   
   CAENVME_FIFOBLTReadCycle(this->handle, 0x32101000, this->blt_buffer,
                         32, cvA32_S_DATA, cvD32, &count);
   printf("count: %d, err: %d\n", count, err_code);

//   this->read_page(8, 0x0000, 0x32110000, cvA32_S_DATA, cvD32, true);
   
   for (unsigned int j=0; j < 8; j++)
      printf("%x\t", this->blt_buffer[j]);
   printf("\n");
   
   return err_code;
}



/////////////////////////////////////////////////////
const char* CAENVMEDevice::decode_error(int err_code)
{  
   switch (err_code)
   {
      case cvSuccess:               return "Operation completed successfully";
      case cvBusError:              return "VME bus error during the cycle";
      case cvCommError:             return "Communication error";
      case cvGenericError:          return "Unspecified error";
      case cvInvalidParam:          return "Invalid parameter";
      case cvTimeoutError:          return "Timeout error";
      case cvAlreadyOpenError:      return "Device is already open error";
      case cvMaxBoardCountError:    return "Maximum device number has been reached";
      case cvNotSupported:          return "Function not supported by that board model";
   }
   return "Unknown error";
}


// aux
////////////////////////////////////////////////////////
void print_params(unsigned char addr_mode=INVALID_MODE,
                  unsigned char data_width=INVALID_MODE,
                  unsigned int base_addr=INVALID_BA)
{
   printf("\t\taddr_mode: 0x%x,\tdata_width: 0x%x,\tbase_addr: 0x%x\n", addr_mode, data_width, base_addr);
}


/////////////////////////////
char* bin(unsigned int value)
{
    char* string = new char[42];
    string[0] = '0';
    string[1] = 'b';
    int i = 2;
    for (int b=31; b>=0; b--) {
       if ((b+1)%4==0) {
          string[i] = '_';
          i++;
       }
       string[i] = ((value>>b) & 1) ? '1' : '0';
       i++;
    }
    string[i] = '\0';
    return string;
}

