#include <stdio.h>
//#include <stdint.h>
//#include <vector.h>
#include <CAENVMEtypes.h>
#ifndef VMELIB_CAENVME_DEVICE_H
#define VMELIB_CAENVME_DEVICE_H

#define INVALID_BA 0xFFFFFFFF
#define INVALID_MODE 0xFF


class CAENVMEDevice
{
   private:   
      unsigned int base_addr;  // 0xXXXX0000
      unsigned int reg_addr;   // 0xXXXX
      unsigned char bd_num;             // board (V1718/V2718) number in CONET interface
      unsigned char link ;              // number of CONET device
      int  handle;             // board handle
      unsigned char bd_type;            // board type
      unsigned char verbosity;
      unsigned char addr_mode;
      unsigned char data_width;
   public:
      int   single_buffer;
      char  string_buffer[100];
      unsigned int blt_buffer[0x4000];

   CAENVMEDevice(unsigned char bd_type,//cvV2718, 
                 unsigned char link,
                 unsigned char bd_num,
                 unsigned char addr_mode,//cvA32_U_DATA,
                 unsigned char data_width,//cvD32,
                 unsigned char verbosity,
                 unsigned int base_addr);

   ~CAENVMEDevice();
   
   int open_board(  unsigned char bd_type,
                    unsigned char link,
                    unsigned char bd_num);

    // prepare VME device (not necessary if base address was set previously)
   int open_device( unsigned int base_addr,
                    unsigned char addr_mode,
                    unsigned char data_width);
// READ / WRITE

// single
   int read_reg(   unsigned int reg_addr,
                   unsigned int base_addr,
                   unsigned char addr_mode,
                   unsigned char data_width);
   
   int write_reg(  unsigned int data,
                   unsigned int reg_addr,
                   unsigned int base_addr,
                   unsigned char addr_mode,
                   unsigned char data_width);
   
   int read_write( unsigned int data,
                   unsigned int reg_addr,
                   unsigned int base_addr,
                   unsigned char addr_mode,
                   unsigned char data_width);

// blocks
   int read_page(  unsigned int size,
                   unsigned int reg_addr,
                   unsigned int base_addr,
                   unsigned char addr_mode,
                   unsigned char data_width,
                   bool  blt);

   int write_page( unsigned int data [],
                   unsigned int size,
                   unsigned int reg_addr,
                   unsigned int base_addr,
                   unsigned char addr_mode,
                   unsigned char data_width,
                   bool  blt);
                   
   int read_multi( unsigned int size,
                   unsigned int regs [],
                   unsigned char addr_modes [],
                   unsigned char data_widths []);

   int write_multi(unsigned int data [],
                   unsigned int size,
                   unsigned int regs [],
                   unsigned char addr_modes [],
                   unsigned char data_widths []);

// IRQ
   int irq_enable(  unsigned int mask);
   
   int irq_disable( unsigned int mask);
   
   int irq_wait(    unsigned int mask,
                    unsigned int timeout);

   int irq_check();

// bus timeout
   int get_timeout();
   int set_timeout(int timeout);

// arbiter type
   int get_arbiter_type();
   int set_arbiter_type(char arbiter_type);

// requester type
   int get_requester_type();
   int set_requester_type(char requester_type);

// requester level
   int get_requester_level();
   int set_requester_level(char requester_level);

// bus release type
   int get_release_type();
   int set_release_type(char release_type);

// reset
//   int reset();
   
   int get_sw_info();
   int test();

// aux   
   const char* decode_error(int err_code);

};

void print_params(unsigned char addr_mode,
                  unsigned char data_width,
                  unsigned int base_addr);
                  
char* bin(unsigned int value);

#endif

