INVALID_BA = 0xffffffff
INVALID_MODE = 0xFF

#Constants from CAENVEtypes.h

# CAEN VME bridges
V1718 = 0                    # CAEN V1718 USB-VME bridge
V2718 = 1                    # V2718 PCI-VME bridge with optical link
A2818 = 2                    # PCI board with optical link
A2719 = 3                    # Optical link piggy-back
A3818 = 4                    # PCIe board with up to 4 optical links

USB_A4818_V2718_LOCAL     = 5
USB_A4818_V2718           = 6
USB_A4818_LOCAL           = 7
USB_A4818_V3718_LOCAL     = 8
USB_A4818_V3718           = 9
USB_A4818_V4718_LOCAL     = 10
USB_A4818_V4718           = 11
USB_A4818                 = 12
USB_A4818_A2719_LOCAL     = 13
USB_V3718_LOCAL           = 14
PCI_A2818_V3718_LOCAL     = 15
PCIE_A3818_V3718_LOCAL    = 16
USB_V3718                 = 17
PCI_A2818_V3718           = 18
PCIE_A3818_V3718          = 19
USB_V4718_LOCAL           = 20
PCI_A2818_V4718_LOCAL     = 21
PCIE_A3818_V4718_LOCAL    = 22
ETH_V4718_LOCAL           = 23
USB_V4718                 = 24
PCI_A2818_V4718           = 25
PCIE_A3818_V4718          = 26
ETH_V4718                 = 27
InvalidBridge             = -1

# VME cycles data width
# Ver. 2.2 - Added byte-swapping data widths
D8  = 0x01                   #  8 bit
D16 = 0x02                   # 16 bit
D32 = 0x04                   # 32 bit
D64 = 0x08                   # 64 bit
D16_swapped = 0x12           # 16 bit swapped
D32_swapped = 0x14           # 32 bit swapped
D64_swapped = 0x18           # 64 bit swapped

# VME cycles address modifiers
A16_S         = 0x2D         # A16 supervisory access
A16_U         = 0x29         # A16 non-privileged
A16_LCK       = 0x2C         # A16 lock command

A24_S_BLT     = 0x3F         # A24 supervisory block transfer
A24_S_PGM     = 0x3E         # A24 supervisory program access
A24_S_DATA    = 0x3D         # A24 supervisory data access
A24_S_MBLT    = 0x3C         # A24 supervisory 64-bit block trnsfer
A24_U_BLT     = 0x3B         # A24 non-privileged block transfer
A24_U_PGM     = 0x3A         # A24 non-privileged program access
A24_U_DATA    = 0x39         # A24 non-privileged data access
A24_U_MBLT    = 0x38         # A24 non-privileged 64-bit block trnsfer
A24_LCK       = 0x32         # A24 lock command

A32_S_BLT     = 0x0F         # A32 supervisory block transfer
A32_S_PGM     = 0x0E         # A32 supervisory program access
A32_S_DATA    = 0x0D         # A32 supervisory data access
A32_S_MBLT    = 0x0C         # A32 supervisory 64-bit block trnsfer
A32_U_BLT     = 0x0B         # A32 non-privileged block transfer
A32_U_PGM     = 0x0A         # A32 non-privileged program access
A32_U_DATA    = 0x09         # A32 non-privileged data access
A32_U_MBLT    = 0x08         # A32 non-privileged 64-bit block trnsfer
A32_LCK       = 0x05         # A32 lock command
CR_CSR        = 0x2F         # CR/CSR space
# The following address modifiers are not yet implemented.
A40_BLT       = 0x37         # A40 block transfer (MD32)
A40_LCK       = 0x35         # A40 lock command
A40           = 0x34         # A40 access
A64           = 0x01         # A64 single trnsfer access
A64_BLT       = 0x03         # A64 block transfer
A64_MBLT      = 0x00         # A64 64-bit block transfer
A64_LCK       = 0x04         # A64 lock command

A3U_2eVME     = 0x21         # 2eVME for 3U modules
A6U_2eVME     = 0x20         # 2eVME for 6U modules

# Error codes returned by the exported functions
Success       =  0           # Operation completed successfully
BusError      = -1           # VME bus error during the cycle  
CommError     = -2           # Communication error             
GenericError  = -3           # Unspecified error               
InvalidParam  = -4           # Invalid parameter               
TimeoutError  = -5           # Timeout error
AlreadyOpenError = -6           # Device is already open error
MaxBoardCountError = -7           # Maximum device number has been reached
NotSupported      = -8            # Function not supported by that board model

# Pulser selection.
PulserA = 0                  # Identifies the pulser 'A'
PulserB = 1                  # Identifies the pulser 'B'

# Output selection.
Output0 = 0                  # Identifies the output line 0
Output1 = 1                  # Identifies the output line 1
Output2 = 2                  # Identifies the output line 2
Output3 = 3                  # Identifies the output line 3
Output4 = 4                  # Identifies the output line 4

# Input selection.
Input0 = 0                   # Identifies the input line 0
Input1 = 1                   # Identifies the input line 1

# Signal sources.
ManualSW    = 0              # Manual (button) or software controlled
InputSrc0   = 1              # Input line 0
InputSrc1   = 2              # Input line 1
Coincidence = 3              # Inputs coincidence
VMESignals  = 4              # Signals from VME bus
MiscSignals = 6              # Various internal signals

# Time base units to specify pulses period and width.
Unit25ns   = 0               # Time unit is 25 nanoseconds
Unit1600ns = 1               # Time unit is 1.6 microseconds
Unit410us  = 2               # Time unit is 410 microseconds
Unit104ms  = 3               # Time unit is 104 milliseconds

# Polarity for LED emitting.
ActiveHigh = 0               # LED emits on signal high level
ActiveLow  = 1               # LED emits on signal low level
# Input and Output signal polarity.
Direct   = 0                 # Normal polarity
Inverted = 1                 # Inverted polarity
# Accessible registers.
StatusReg      = 0x00        # Status register
VMEControlReg  = 0x01        # VME Control register
FwRelReg       = 0x02        # Firmware Release register
FwDldReg       = 0x03        # Firmware Download register
FlenaReg       = 0x04        # Flash Enable
VMEIRQEnaReg   = 0x06        # VME IRQ Lines Enable
InputReg       = 0x08        # Input register
OutRegSet      = 0x0A        # Output register
InMuxRegSet    = 0x0B        # Input Multiplexer
OutMuxRegSet   = 0x0C        # Output Multiplexer
LedPolRegSet   = 0x0D        # Led Polarity
OutRegClear    = 0x10        # Output register
InMuxRegClear  = 0x11        # Input Multiplexer
OutMuxRegClear = 0x12        # Output Multiplexer
LedPolRegClear = 0x13        # Led Polarity
PulserA0       = 0x16        # Period and width of Pulser A
PulserA1       = 0x17        # Num pulses and range of Pulser A
PulserB0       = 0x19        # Period and width of Pulser B
PulserB1       = 0x1A        # Num pulses and range of Pulser B
Scaler0        = 0x1C        # Limit and Autores of Scaler A
Scaler1        = 0x1D        # Counter value of Scaler A
DispADL        = 0x20        # Display AD[15:0]
DispADH        = 0x21        # Display AD[31:16]
DispDTL        = 0x22        # Display DT[15:0]
DispDTH        = 0x23        # Display DT[31:16]
DispC1         = 0x24        # Display Control left bar
DispC2         = 0x25        # Display Control left bar
LMADL          = 0x28        # Loc. Mon. AD[15:0]
LMADH          = 0x29        # Loc. Mon. AD[31:16]
LMC            = 0x2C        # Loc. Mon. Controls
# Bits for status register decoding.
SYSRES   = 0x0001            # VME is in system reset state
SYSCTRL  = 0x0002            # The bridge is the VME system controller
DTACK    = 0x0010            # Last access has generated a DTACK signal
BERR     = 0x0020            # Last access has generated a bus error
DIP0     = 0x0100            # Dip Switch position 0 state
DIP1     = 0x0200            # Dip Switch position 1 state
DIP2     = 0x0400            # Dip Switch position 2 state
DIP3     = 0x0800            # Dip Switch position 3 state
DIP4     = 0x1000            # Dip Switch position 4 state
USBTYPE  = 0x8000            # USB Speed: 0 = Full; 1 = High
# Bits for input register decoding.
In0Bit        = 0x0001       # Input line 0 signal level.
In1Bit        = 0x0002       # Input line 1 signal level.
CoincBit      = 0x0004       # Coincidence of input signal level.
PulsAOutBit   = 0x0008       # Pulser A output signal level.
PulsBOutBit   = 0x0010       # Pulser B output signal level.
ScalEndCntBit = 0x0020       # Scaler end counter signal level.
LocMonBit     = 0x0040       # Location monitor signal level.
# Bits for input register decoding.
PulsAStartBit = 0x0001       # Pulser A start signal level.
PulsAResetBit = 0x0002       # Pulser A reset signal level.
PulsBStartBit = 0x0004       # Pulser B start signal level.
PulsBResetBit = 0x0008       # Pulser B reset signal level.
ScalGateBit   = 0x0010       # Scaler gate signal level.
ScalResetBit  = 0x0020       # Scaler reset counter signal level.
Out0Bit       = 0x0040       # Output line 0 signal level.
Out1Bit       = 0x0080       # Output line 1 signal level.
Out2Bit       = 0x0100       # Output line 2 signal level.
Out3Bit       = 0x0200       # Output line 3 signal level.
Out4Bit       = 0x0400       # Output line 4 signal level.
# Types of VME Arbiter.
ArbiterPriorized  = 0               # Priority Arbiter
ArbiterRoundRobin = 1               # Round-Robin Arbiter
# Types of VME Bus Requester.
RequesterFair   = 0                   # Fair bus requester
RequesterDemand = 1                   # On demand bus requester
# Types of VME Bus release.
ReleaseRWD   = 0                    # Release When Done
ReleaseROR   = 1                    # Release On Request
# VME bus request levels.
BR0 = 0                      # Bus request level 0
BR1 = 1                      # Bus request level 1
BR2 = 2                      # Bus request level 2
BR3 = 3                      # Bus request level 3
# VME Interrupt levels.
IRQ1 = 0x01                  # Interrupt level 1
IRQ2 = 0x02                  # Interrupt level 2
IRQ3 = 0x04                  # Interrupt level 3
IRQ4 = 0x08                  # Interrupt level 4
IRQ5 = 0x10                  # Interrupt level 5
IRQ6 = 0x20                  # Interrupt level 6
IRQ7 = 0x40                  # Interrupt level 7
# VME bus timeouts.
Timeout50us   = 0            # Timeout is 50 microseconds
Timeout400us  = 1            #  Timeout is 400 microseconds

