#include <Python.h>
#include <vmelib.h>
#include <numpy/arrayobject.h>
#include <stdio.h>
#include <stdint.h>


///////////////////////////////////////////////////////////////////////
PyObject* construct(PyObject* self, PyObject* args)
{
    // Arguments passed from Python
    unsigned char bd_type_;         //cvV2718, 
    unsigned char link_; 
    unsigned char bd_num_;
    unsigned char addr_mode_;       //cvA32_U_DATA,
    unsigned char data_width_;      //cvD32,
    unsigned char verbosity_;
    unsigned int base_addr_;
    Py_Initialize();
    _import_array();
    
    // Process arguments passes from Python
    PyArg_ParseTuple(args, "bbbbbbI",
                        &bd_type_,
                        &link_,
                        &bd_num_,
                        &addr_mode_,
                        &data_width_,
                        &verbosity_,
                        &base_addr_);

    // Allocate new `CAENVMEDevice` object
   CAENVMEDevice* device = new CAENVMEDevice(    bd_type_,
                                                link_,
                                                bd_num_,
                                                addr_mode_,
                                                data_width_,
                                                verbosity_,
                                                base_addr_);

    // Create a Python capsule with a pointer to `CAENVMEDevice` object
    PyObject* vme_device = PyCapsule_New((void *)device, "dev_ptr", NULL);
    PyCapsule_SetPointer(vme_device, (void *)device);

    // Return the Python capsule with the pointer to`CAENVMEDevice` object
    return Py_BuildValue("O", vme_device);   // "O" means "Python object"
}


///////////////////////////////////////////////////////////////////////
PyObject* delete_object(PyObject* self, PyObject* args)
{
//  C++/PYTHON `CAENVMEDevice` OBJECT DESTRUCTOR.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `vme_device` object

    // Process arguments
    PyArg_ParseTuple(args, "O", &vme_device_capsule_);

    // Get the pointer to `vme_device_ptr` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");

    // Delete the `CAENVMEDevice` object
    delete device;

    // Return nothing
    return Py_BuildValue("");
}


///////////////////////////////////////////////////////////////////////
PyObject* open_board(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::open_board.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object
    unsigned char bd_type_;
    unsigned char link_;
    unsigned char bd_num_;

    // Process arguments
    PyArg_ParseTuple(   args, "Obbb",
                        &vme_device_capsule_,
                        &bd_type_,
                        &link_,
                        &bd_num_);

    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");

    // Call `open_board` function
       device->open_board(bd_type_, link_, bd_num_);

    // Return error code
    return Py_BuildValue("i", cvSuccess);    
}

///////////////////////////////////////////////////////////////////////
PyObject* open_device(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::open_device.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object
    unsigned int base_addr_;
    unsigned char addr_mode_;
    unsigned char data_width_;

    // Process arguments
    PyArg_ParseTuple(args, "OIbb",
                    &vme_device_capsule_,
                    &base_addr_,
                    &addr_mode_,
                    &data_width_);

    // Get the pointer to `vme_device` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");

    // Call `open_device` function
    int err_code = device->open_device(base_addr_, addr_mode_, data_width_);

    // Return error code
    return Py_BuildValue("i", err_code);
}

///////////////////////////////////////////////////////////////////////
PyObject* read_reg(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::read_reg.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object
    unsigned int reg_addr_;
    unsigned int base_addr_;
    unsigned char addr_mode_;
    unsigned char data_width_;

    // Process arguments
    PyArg_ParseTuple(args, "OIIbb",
                    &vme_device_capsule_,
                    &reg_addr_,
                    &base_addr_,
                    &addr_mode_,
                    &data_width_);

    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");

    // Call `read_reg` function
    int err_code = device->read_reg(reg_addr_, base_addr_, addr_mode_, data_width_);
    unsigned int data = device->single_buffer;
    // Return data from the device
    return Py_BuildValue("iI", err_code, data);
}


///////////////////////////////////////////////////////////////////////
PyObject* write_reg(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::write_reg.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object
    unsigned int data_;
    unsigned int reg_addr_;
    unsigned int base_addr_;
    unsigned char addr_mode_;
    unsigned char data_width_;

    // Process arguments
    PyArg_ParseTuple(args, "OIIIbb",
                    &vme_device_capsule_,
                    &data_,
                    &reg_addr_,
                    &base_addr_,
                    &addr_mode_,
                    &data_width_);
    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");

    // Call `write_reg` function
    int err_code = device->write_reg(data_, reg_addr_, base_addr_, addr_mode_, data_width_);

    // Return buffern contain
    return Py_BuildValue("i", err_code);
}


///////////////////////////////////////////////////////////////////////
PyObject* read_write(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::read_write.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object
    unsigned int data_;
    unsigned int reg_addr_;
    unsigned int base_addr_;
    unsigned char addr_mode_;
    unsigned char data_width_;

    // Process arguments
    PyArg_ParseTuple(args, "OIIIbb",
                    &vme_device_capsule_,
                    &data_,
                    &reg_addr_,
                    &base_addr_,
                    &addr_mode_,
                    &data_width_);
    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");

    // Call `write_reg` function
    int err_code = device->read_write(data_, reg_addr_, base_addr_, addr_mode_, data_width_);

    // Return buffern contain
    return Py_BuildValue("iI", err_code, device->single_buffer);
}


///////////////////////////////////////////////////////////////////////
PyObject* read_page(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::read_page.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object
    unsigned int size_;
    unsigned int reg_addr_;
    unsigned int base_addr_;
    unsigned char addr_mode_;
    unsigned char data_width_;
    bool blt_;

    // Process arguments
    PyArg_ParseTuple(args, "OIIIbbb",
                    &vme_device_capsule_,
                    &size_,
                    &reg_addr_,
                    &base_addr_,
                    &addr_mode_,
                    &data_width_,
                    &blt_);

    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");

    // Call `read_page` function
    int err_code = device->read_page(size_, reg_addr_, base_addr_, addr_mode_, data_width_, blt_); //buf_size is the number of 4-byte words
    npy_intp arr_shape[1] = { device->single_buffer };
    unsigned int* p_bufer = device->blt_buffer;
    PyObject *out_array;
    out_array = PyArray_SimpleNewFromData(1, arr_shape, NPY_UINT32, (void*)p_bufer);    

    return Py_BuildValue("iO", err_code, out_array);
}


///////////////////////////////////////////////////////////////////////
PyObject* write_page(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::write_page.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object
    unsigned int size_;
    unsigned int reg_addr_;
    unsigned int base_addr_;
    unsigned char addr_mode_;
    unsigned char data_width_;
    bool blt_;
    PyObject * data;

    // Process arguments
    PyArg_ParseTuple(args, "OOIIbbb",
                    &vme_device_capsule_,
                    &data,
                    &reg_addr_,
                    &base_addr_,
                    &addr_mode_,
                    &data_width_,
                    &blt_);
    size_ = PyList_Size(data);
    unsigned int * data_ = new unsigned int [size_];
    PyObject * py_obj;

    // get data from list
    for (unsigned int i=0; i<size_; i++) {
        py_obj = PyList_GetItem(data, i);
        if (!PyLong_Check(py_obj)) {
            PyErr_SetString(PyExc_TypeError, "list items must be integers.");
            return NULL;
        }
        data_[i] = PyLong_AsLong(py_obj) & 0xffffffff;
    }

    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");

    // Call `write_page` function
    int err_code = device->write_page(data_, size_, reg_addr_, base_addr_, addr_mode_, data_width_, blt_);
    
    // Return buffern contain
    return Py_BuildValue("i", err_code);
}


///////////////////////////////////////////////////////////////////////
PyObject* read_multi(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::read_reg.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object
    PyObject     * regs;
    PyObject * addr_modes;
    PyObject * data_widths;
    // C objects
    unsigned int size_;
    PyObject * py_obj;
    // Process arguments
    PyArg_ParseTuple(args, "OOOO",
                     &vme_device_capsule_,
                     &regs,
                     &addr_modes,
                     &data_widths);
    // size of arrays by len(regs)
    size_ = PyList_Size(regs);
    unsigned int * regs_ = new unsigned int [size_];
    unsigned char  *  addr_modes_ = new unsigned char [size_];
    unsigned char  * data_widths_ = new unsigned char [size_];

    for (unsigned int i=0; i<size_; i++) {
        // regs
        py_obj = PyList_GetItem(regs, i);
        if (!PyLong_Check(py_obj)) {
            PyErr_SetString(PyExc_TypeError, "list items must be integers.");
            return NULL;
        }
        regs_[i] = PyLong_AsLong(py_obj) & 0xffffffff;
        // addr_modes
        py_obj = PyList_GetItem(addr_modes, i);
        if (!PyLong_Check(py_obj)) {
            PyErr_SetString(PyExc_TypeError, "list items must be integers.");
            return NULL;
        }
        addr_modes_[i] = PyLong_AsLong(py_obj) & 0xff;
        // data_widths
        py_obj = PyList_GetItem(data_widths, i);
        if (!PyLong_Check(py_obj)) {
            PyErr_SetString(PyExc_TypeError, "list items must be integers.");
            return NULL;
        }
        data_widths_[i] = PyLong_AsLong(py_obj) & 0xff;
    }

    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");

    // Call `read_page` function
    int err_code = device->read_multi(size_, regs_, addr_modes_, data_widths_); //buf_size is the number of 4-byte words
    npy_intp arr_shape[1] = { device->single_buffer };
    unsigned int* p_bufer = device->blt_buffer;
    PyObject *out_array;
    out_array = PyArray_SimpleNewFromData(1, arr_shape, NPY_UINT32, (void*)p_bufer);    
    return Py_BuildValue("iO", err_code, out_array);
}


///////////////////////////////////////////////////////////////////////
PyObject* write_multi(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::read_reg.
    // Arguments passed from Python
    PyObject * vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object
    PyObject * data;
    PyObject * regs;
    PyObject * addr_modes;
    PyObject * data_widths;
    // C objects
    unsigned int size_;
    PyObject * py_obj;
    // Process arguments
    PyArg_ParseTuple(args, "OOOOO",
                     &vme_device_capsule_,
                     &data,
                     &regs,
                     &addr_modes,
                     &data_widths);
    // size of arrays by len(regs)
    size_ = PyList_Size(regs);
    unsigned int  * data_ = new unsigned int [size_];
    unsigned int  * regs_ = new unsigned int [size_];
    unsigned char * addr_modes_ = new unsigned char [size_];
    unsigned char * data_widths_ = new unsigned char [size_];

    for (unsigned int i=0; i<size_; i++) {
        // data
        py_obj = PyList_GetItem(data, i);
        if (!PyLong_Check(py_obj)) {
            PyErr_SetString(PyExc_TypeError, "list items must be integers.");
            return NULL;
        }
        data_[i] = PyLong_AsLong(py_obj) & 0xffffffff;
        // regs
        py_obj = PyList_GetItem(regs, i);
        if (!PyLong_Check(py_obj)) {
            PyErr_SetString(PyExc_TypeError, "list items must be integers.");
            return NULL;
        }
        regs_[i] = PyLong_AsLong(py_obj) & 0xffffffff;
        // addr_modes
        py_obj = PyList_GetItem(addr_modes, i);
        if (!PyLong_Check(py_obj)) {
            PyErr_SetString(PyExc_TypeError, "list items must be integers.");
            return NULL;
        }
        addr_modes_[i] = PyLong_AsLong(py_obj) & 0xff;
        // data_widths
        py_obj = PyList_GetItem(data_widths, i);
        if (!PyLong_Check(py_obj)) {
            PyErr_SetString(PyExc_TypeError, "list items must be integers.");
            return NULL;
        }
        data_widths_[i] = PyLong_AsLong(py_obj) & 0xff;
    }

    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");

    // Call `write_multi` function
    int err_code = device->write_multi(data_, size_, regs_, addr_modes_, data_widths_);
//    npy_intp arr_shape[1] = { device->single_buffer }; // return also a number of written bytes?
    return Py_BuildValue("i", err_code);
}


///////////////////////////////////////////////////////////////////////
PyObject* irq_enable(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::irq_enable.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object
    unsigned int mask_;

    // Process arguments
    PyArg_ParseTuple(args, "OI",
                    &vme_device_capsule_,
                    &mask_);
    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");

    // Call `irq_enable` function
    int err_code = device->irq_enable(mask_);

    // Return buffern contain
    return Py_BuildValue("i", err_code);
}


///////////////////////////////////////////////////////////////////////
PyObject* irq_disable(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::irq_disable.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object
    unsigned int mask_;

    // Process arguments
    PyArg_ParseTuple(args, "OI",
                    &vme_device_capsule_,
                    &mask_);
    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");

    // Call `irq_disable` function
    int err_code = device->irq_disable(mask_);

    // Return buffern contain
    return Py_BuildValue("i", err_code);
}


///////////////////////////////////////////////////////////////////////
PyObject* irq_wait(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::irq_wait.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object
    unsigned int mask_;
    unsigned int timeout_;

    // Process arguments
    PyArg_ParseTuple(args, "OII",
                    &vme_device_capsule_,
                    &mask_,
                    &timeout_);
    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");

    // Call `irq_wait` function
    int err_code = device->irq_wait(mask_, timeout_);

    return Py_BuildValue("i", err_code);
}


///////////////////////////////////////////////////////////////////////
PyObject* irq_check(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::irq_check.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object

    // Process arguments
    PyArg_ParseTuple(args, "O",
                    &vme_device_capsule_);
    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");

    // Call `irq_check` function
    int err_code = device->irq_check();

    // Return buffer contain
    return Py_BuildValue("ii", err_code, device->single_buffer);
}


///////////////////////////////////////////////////////////////////////
PyObject* get_timeout(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::get_timeout.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object

    // Process arguments
    PyArg_ParseTuple(args, "O",
                    &vme_device_capsule_);
    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");

    // Call `get_timeout` function
    int err_code = device->get_timeout();

    // Return buffer contain
    return Py_BuildValue("ii", err_code, device->single_buffer);
}


///////////////////////////////////////////////////////////////////////
PyObject* set_timeout(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::set_timeout.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object
    int timeout_;

    // Process arguments
    PyArg_ParseTuple(args, "Oi",
                    &vme_device_capsule_,
                    &timeout_);
    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");

    // Call `set_timeout` function
    int err_code = device->set_timeout(timeout_);

    // Return buffer content
    return Py_BuildValue("i", err_code);
}


///////////////////////////////////////////////////////////////////////
PyObject* get_arbiter_type(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::get_arbiter_type.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object

    // Process arguments
    PyArg_ParseTuple(args, "O",
                    &vme_device_capsule_);
    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");
    // Call `get_arbiter_type` function
    int err_code = device->get_arbiter_type();

    // Return buffer contain
    return Py_BuildValue("ii", err_code, device->single_buffer);
//    return Py_BuildValue("ii", 0, 0);
}


///////////////////////////////////////////////////////////////////////
PyObject* set_arbiter_type(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::set_arbiter_type.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object
    int arbiter_type_;

    // Process arguments
    PyArg_ParseTuple(args, "Oi",
                    &vme_device_capsule_,
                    &arbiter_type_);
    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");

    // Call `set_arbiter_type` function
    int err_code = device->set_arbiter_type(arbiter_type_);

    // Return buffer content
    return Py_BuildValue("i", err_code);
}


///////////////////////////////////////////////////////////////////////
PyObject* get_requester_type(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::get_requester_type.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object

    // Process arguments
    PyArg_ParseTuple(args, "O",
                    &vme_device_capsule_);
    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");
    // Call `get_requester_type` function
    int err_code = device->get_requester_type();

    // Return buffer content
    return Py_BuildValue("ii", err_code, device->single_buffer);
}


///////////////////////////////////////////////////////////////////////
PyObject* set_requester_type(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::set_requester_type.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object
    int requester_type_;

    // Process arguments
    PyArg_ParseTuple(args, "Oi",
                    &vme_device_capsule_,
                    &requester_type_);
    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");

    // Call `set_requester_type` function
    int err_code = device->set_requester_type(requester_type_);

    // Return buffer content
    return Py_BuildValue("i", err_code);
}


///////////////////////////////////////////////////////////////////////
PyObject* get_requester_level(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::get_requester_level.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object

    // Process arguments
    PyArg_ParseTuple(args, "O",
                    &vme_device_capsule_);
    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");
    // Call `get_requester_level` function
    int err_code = device->get_requester_level();

    // Return buffer content
    return Py_BuildValue("ii", err_code, device->single_buffer);
}


///////////////////////////////////////////////////////////////////////
PyObject* set_requester_level(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::set_requester_level.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object
    int requester_level_;

    // Process arguments
    PyArg_ParseTuple(args, "Oi",
                    &vme_device_capsule_,
                    &requester_level_);
    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");

    // Call `set_arbiter_type` function
    int err_code = device->set_requester_level(requester_level_);

    // Return buffer content
    return Py_BuildValue("i", err_code);
}


///////////////////////////////////////////////////////////////////////
PyObject* get_release_type(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::get_release_type.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object

    // Process arguments
    PyArg_ParseTuple(args, "O",
                    &vme_device_capsule_);
    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");
    // Call `get_requester_type` function
    int err_code = device->get_release_type();

    // Return buffer content
    return Py_BuildValue("ii", err_code, device->single_buffer);
}


///////////////////////////////////////////////////////////////////////
PyObject* set_release_type(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::set_release_type.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object
    int release_type_;

    // Process arguments
    PyArg_ParseTuple(args, "Oi",
                    &vme_device_capsule_,
                    &release_type_);
    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");

    // Call `set_requester_type` function
    int err_code = device->set_release_type(release_type_);

    // Return buffer content
    return Py_BuildValue("i", err_code);
}


///////////////////////////////////////////////////////////////////////
PyObject* get_sw_info(PyObject* self, PyObject* args)
{
//  C++/PYTHON WRAPPER FOR PUBLIC FUNCTION CAENVMEDevice::get_sw_info.
    // Arguments passed from Python
    PyObject* vme_device_capsule_;   // Capsule with the pointer to `CAENVMEDevice` object

    // Process arguments
    PyArg_ParseTuple(args, "O",
                    &vme_device_capsule_);
    // Get the pointer to `CAENVMEDevice` object
    CAENVMEDevice* device = (CAENVMEDevice*)PyCapsule_GetPointer(vme_device_capsule_, "dev_ptr");

    // Call `irq_check` function
    int err_code = device->get_sw_info();

    // Return buffer content
    return Py_BuildValue("is", err_code, device->string_buffer);
}


///////////////////////////////////////////////////////////////////////
PyMethodDef CCAENVMEDevice_functions[] =
{
//  Structures which define functions ("methods") provided by the module.
    {"construct",                   // C++/Py Constructor
     construct, METH_VARARGS,
     "Create CAENVMEDevice"},

    {"delete_object",               // C++/Py Destructor
     delete_object, METH_VARARGS,
     "Delete CAENVMEDevice"},

    // C++/Py wrappers for CAENVMEDevice methods
    {"open_board",
     open_board, METH_VARARGS,
     "Open main board for connection to VME bus"},

    {"open_device",
     open_device, METH_VARARGS,
     "Open device connected to VME bus"},

    {"read_reg",
     read_reg, METH_VARARGS,
     "Read register from the device on VME bus"},
    
    {"write_reg",
     write_reg, METH_VARARGS,
     "Write to the register of the device on VME bus"},

    {"read_write",
     read_write, METH_VARARGS,
     "Read and then write to the register of the device on VME bus"},

    {"read_page",
     read_page, METH_VARARGS,
     "Read an array of contiguous registers from the device on VME bus"},

    {"write_page",
     write_page, METH_VARARGS,
     "Write an array of contiguous registers from the device on VME bus"},

    {"read_multi",
     read_multi, METH_VARARGS,
     "Read an array of arbitrary registers from the device on VME bus"},
     
    {"write_multi",
     write_multi, METH_VARARGS,
     "Read an array of arbitrary registers from the device on VME bus"},
     
    {"irq_enable",
     irq_enable, METH_VARARGS,
     "Enable IRQ with mask"},

    {"irq_disable",
     irq_disable, METH_VARARGS,
     "Disable IRQ with mask"},

    {"irq_wait",
     irq_wait, METH_VARARGS,
     "Recieve an interrupt with mask or return timeout error"},

    {"irq_check",
     irq_check, METH_VARARGS,
     "Check interrupt state"},

    {"get_timeout",
     get_timeout, METH_VARARGS,
     "Recieve VME bus timeout"},
     
    {"set_timeout",
     set_timeout, METH_VARARGS,
     "Set VME bus timeout"},

    {"get_arbiter_type",
     get_arbiter_type, METH_VARARGS,
     "Get the type of VME bus arbiter implemented on the module"},
     
    {"set_arbiter_type",
     set_arbiter_type, METH_VARARGS,
     "Set the type of VME bus arbiter implemented on the module"},
     
    {"get_requester_type",
     get_requester_type, METH_VARARGS,
     "Get VME bus requester type"},
     
    {"set_requester_type",
     set_requester_type, METH_VARARGS,
     "Set VME bus requester type"},
     
    {"get_requester_level",
     get_requester_level, METH_VARARGS,
     "Get VME bus requester priority level"},
     
    {"set_requester_level",
     set_requester_level, METH_VARARGS,
     "Set VME bus requester priority level"},
     
    {"get_release_type",
     get_release_type, METH_VARARGS,
     "Get VME bus release type"},
     
    {"set_release_type",
     set_release_type, METH_VARARGS,
     "Set VME bus release type"},
          
    {"get_sw_info",
     get_sw_info, METH_VARARGS,
     "Get versions of the library and board firmware"},
     
    {NULL, NULL, 0, NULL}   /* Last function description must be empty.
                               Otherwise, it will create seg fault while
                               importing the module. */
};

struct PyModuleDef CCAENVMEDevice =
{
/*  Structure which defines the module.
 *  For more info look at: https://docs.python.org/3/c-api/module.html
 */
   PyModuleDef_HEAD_INIT,
   "CCAENVMEDevice",         // Name of the module.
   "C++ to Python wrapper",  // Docstring for the module.
   -1,                       // Used by sub-interpreters, if you do not know what
                             // it is then you do not need it, keep -1 .
   CCAENVMEDevice_functions  // Structures of type `PyMethodDef` with functions
                             // (or "methods") provided by the module.
};

PyMODINIT_FUNC PyInit_CCAENVMEDevice(void)
{
/*
 *   Function which initialises the Python module.
 *
 *   Note:  This function must be named "PyInit_MODULENAME",
 *          where "MODULENAME" is the name of the module.
 *
 */
    return PyModule_Create(&CCAENVMEDevice);
}

